# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = 'pymeasurement'
copyright = '2023, Saptak Das'
author = 'Saptak Das'
release = '1.0.9'

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

import os
import sys

sys.path.append(os.path.abspath("./sphinxcontrib"))

extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.viewcode',
    'sphinx.ext.githubpages',
    'sphinx.ext.doctest',
    'sphinx.ext.mathjax',
    'sphinx_copybutton',
    'sphinxcontrib.exceltable'
]

sys.path.insert(0, os.path.abspath('../../src/'))

templates_path = ['_templates']
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']

html_theme = 'furo'
html_static_path = ['_static']



from decimal import Decimal, Context

class SigFig:
  """A class for representing numbers with significant figures. SigFig objects are immutable. Internally, all numbers are stored as Decimal objects (fixed point numbers) for extra accuracy and precision. The central paradigm of this class is that the decimal value is the true value of the number, and the sigfigs and decimals are the precision of the number. The sigfigs and decimals are used to determine the precision of the number when it is printed.

  :param value: The value of the number as a string.
  :type value: str
  :param sigfigs: The number of significant figures to use when printing the number. If None, the number of significant figures will be automatically determined.
  :type sigfigs: int or None
  :param decimals: The number of decimal places to use when printing the number. If None, the number of decimal places will be automatically determined.
  :type decimals: int or None
  :param constant: If True, the number will be assumed to be perfectly accurate for all calculations. If False, the number will be assumed to have some precision that must be followed.
  :type constant: bool
  """
  def __init__(self, value, sigfigs=None, decimals=None, constant=False):
    """SigFig Constructor
    """
    self.value = value
    try:
      self.decimalValue = Decimal(value) #True Value of Decimal including extra calculation precision.
    except:
      raise Exception(f'Sig Fig Error: Could not convert "{self.value}" into sig fig.')
    sign, digits, exponent = self.decimalValue.as_tuple()
    self.sigfigs = len(digits)
    self.decimal = Decimal((sign, digits, exponent)) #Sig Fig Decimal Representation
    if constant:
      #Constants are assumed to be perfectly accurate for all calculations.
      self.sigfigs = float('inf')
      self.decimals = float('-inf')
    else:
      #Value has some precision that must be followed.
      #Automatic Override
      if '.' in self.value: #Decimal Value
        #Force override to maintain sig fig precision
        self.decimal = SigFig.changeSigFigs(value, self.sigfigs)
        self.decimals = exponent
      else:
        newSigfigs = len([int(i) for i in ''.join([str(i) for i in digits]).rstrip('0')])
        if newSigfigs != self.sigfigs and newSigfigs > 0:
          self.decimal = SigFig.changeSigFigs(value, newSigfigs)
          self.sigfigs = newSigfigs
        self.decimals = len(digits) - self.sigfigs
  
      #Manual override for sigfig or decimal precision.
      if sigfigs != None and sigfigs != float("inf"):
        self.sigfigs = sigfigs
        self.decimal = SigFig.changeSigFigs(value, sigfigs)
        sign, digits, exponent = self.decimal.as_tuple()
        if exponent < 0: #Decimal Value
          self.decimals = exponent
        else:
          self.decimals = len(digits) - self.sigfigs
      elif decimals != None:
        self.decimals = decimals
        self.decimal = self.decimal.quantize(Decimal(f"1E{self.decimals}"))
        sign, digits, exponent = self.decimal.as_tuple()
        self.sigfigs = len(digits) - self.decimals + exponent

  def changeSigFigs(value, sigfigs):
    """Changes the number of significant figures of a number.

    :param value: The value of the number as a string.
    :type value: str
    :param sigfigs: The number of significant figures to use.
    :type sigfigs: int
    :return: The number with the new number of significant figures.
    :rtype: Decimal
    """
    sign, digits, exponent = Context(prec=sigfigs).create_decimal(value).as_tuple()
    if len(digits) < sigfigs:
      missing = sigfigs - len(digits)
      digits = digits + (0,) * missing
      exponent -= missing
    return Decimal((sign, digits, exponent))

  def deepCopy(self):
    """Returns a deep copy of the SigFig object.
    
    :return: A deep copy of the SigFig object.
    :rtype: SigFig
    """
    new = SigFig('0')
    new.value = self.value
    new.decimalValue = self.decimalValue
    new.decimal = self.decimal
    new.sigfigs = self.sigfigs
    new.decimals = self.decimals
    return new
  
  def __str__(self):
    """Returns the string representation of the SigFig object.

    :return: The string representation of the SigFig object.
    :rtype: str
    """
    return str(self.decimal)

  def __repr__(self):
    """Returns the string representation of the SigFig object.

    :return: The string representation of the SigFig object.
    :rtype: str
    """
    return str(self)

  def __eq__(self, other):
    """Checks if the two SigFig objects are equal.

    :param other: The other SigFig object to compare to this one.
    :type other: SigFig
    :return: True if the two SigFig objects are equal.
    :rtype: bool
    """
    return self.decimal == other.decimal

  def __lt__(self, other):
    """Checks if the SigFig object is less than the other SigFig object.

    :param other: The other SigFig object to compare to this one.
    :type other: SigFig
    :return: True if the SigFig object is less than the other SigFig object.
    :rtype: bool
    """
    return self.decimal < other.decimal

  def __gt__(self, other):
    """Checks if the SigFig object is greater than the other SigFig object.

    :param other: The other SigFig object to compare to this one.
    :type other: SigFig
    :return: True if the SigFig object is greater than the other SigFig object.
    :rtype: bool
    """
    return self.decimal > other.decimal

  def __le__(self, other):
    """Checks if the SigFig object is less than or equal to the other SigFig object.

    :param other: The other SigFig object to compare to this one.
    :type other: SigFig
    :return: True if the SigFig object is less than or equal to the other SigFig object.
    :rtype: bool
    """
    return self < other or self == other

  def __ge__(self, other):
    """Checks if the SigFig object is greater than or equal to the other SigFig object.

    :param other: The other SigFig object to compare to this one.
    :type other: SigFig
    :return: True if the SigFig object is greater than or equal to the other SigFig object.
    :rtype: bool
    """
    return self > other or self == other
  
  def __neg__(self):
    """Returns the negative of the SigFig object.

    :return: The negative of the SigFig object.
    :rtype: SigFig
    """
    neg = self.deepCopy()
    neg.value = self.value.replace('-', '') if '-' in self.value else f'-{self.value}'
    neg.decimal = -self.decimal
    neg.decimalValue = -self.decimalValue
    return neg

  def __add__(self, other):
    """Returns the sum of the two SigFig objects as a new SigFig object, following the rules of significant figures.

    :param other: The other SigFig object to add to this one.
    :type other: SigFig
    :return: The sum of the two SigFig objects as a new SigFig object.
    :rtype: SigFig
    """
    if isinstance(other, float) or isinstance(other, int):
      other = SigFig(str(other), constant=True)
    decimals = max(self.decimals, other.decimals)
    return SigFig(str(self.decimalValue + other.decimalValue), decimals=decimals, constant=decimals == float('-inf'))
  
  def __radd__(self, other):
    """Returns the sum of the two SigFig objects as a new SigFig object, following the rules of significant figures.

    :param other: The other SigFig object to add to this one.
    :type other: SigFig
    :return: The sum of the two SigFig objects as a new SigFig object.
    :rtype: SigFig
    """
    return self + other
  
  def __sub__(self, other):
    """Returns the difference of the two SigFig objects as a new SigFig object, following the rules of significant figures.

    :param other: The other SigFig object to subtract from this one.
    :type other: SigFig
    :return: The difference of the two SigFig objects as a new SigFig object.
    :rtype: SigFig
    """
    if isinstance(other, float) or isinstance(other, int):
      other = SigFig(str(other), constant=True)
    return -other + self

  def __rsub__(self, other):
    """Returns the difference of the two SigFig objects as a new SigFig object, following the rules of significant figures.

    :param other: The other SigFig object to subtract from this one.
    :type other: SigFig
    :return: The difference of the two SigFig objects as a new SigFig object.
    :rtype: SigFig
    """
    return -self + other

  def __mul__(self, other):
    """Returns the product of the two SigFig objects as a new SigFig object, following the rules of significant figures.

    :param other: The other SigFig object to multiply by this one.
    :type other: SigFig
    :return: The product of the two SigFig objects as a new SigFig object.
    :rtype: SigFig
    """
    if isinstance(other, float) or isinstance(other, int):
      other = SigFig(str(other), constant=True)
    sigfigs = min(self.sigfigs, other.sigfigs)
    return SigFig(str(self.decimalValue * other.decimalValue), sigfigs=sigfigs, constant=sigfigs == float('inf'))

  def __rmul__(self, other):
    """Returns the product of the two SigFig objects as a new SigFig object, following the rules of significant figures.

    :param other: The other SigFig object to multiply by this one.
    :type other: SigFig
    :return: The product of the two SigFig objects as a new SigFig object.
    :rtype: SigFig
    """
    return self * other

  def __truediv__(self, other):
    """Returns the quotient of the two SigFig objects as a new SigFig object, following the rules of significant figures.

    :param other: The other SigFig object to divide by this one.
    :type other: SigFig
    :return: This SigFig object divided by the other SigFig object.
    :rtype: SigFig
    """
    if isinstance(other, float) or isinstance(other, int):
      other = SigFig(str(other), constant=True)
    sigfigs = min(self.sigfigs, other.sigfigs)
    return SigFig(str(self.decimalValue / other.decimalValue), sigfigs=sigfigs, constant=sigfigs == float('inf'))

  def __rtruediv__(self, other):
    """Returns the quotient of the two SigFig objects as a new SigFig object, following the rules of significant figures.

    :param other: The other SigFig object to divide by this one.
    :type other: SigFig
    :return: Other SigFig object divided by this one.
    :rtype: SigFig
    """
    return other / self

  def abs(self):
    """Returns the absolute value of the SigFig object.

    :return: The absolute value of the SigFig object.
    :rtype: SigFig
    """
    if self >= SigFig('0', constant=True):
      return self
    else:
      return -self


from pymeasurement.util.parser import Parser
from pymeasurement.util.chem.element import Element
from pymeasurement.measurement import Measurement
import re

class Compound(Parser):
  """A class to represent a chemical compound.

  :param string: The string to parse.
  :type string: str
  """

  latexPrint = False # Whether to print the compound in latex format.

  def setLatexPrint(value):
    """Set whether to print the compound in latex format.

    :param value: Whether to print the compound in latex format.
    :type value: bool
    """
    Element.setLatexPrint(value)
    Compound.latexPrint = value
  
  def __init__(self, string):
    """Compound Constructor
    """
    super().__init__(string)
    self.composition = {}
    self.element = ""
    self.number = ""
    self.parenthesesOn = False
    self.parentheses = ')'
    self.subString = ""
    self.compoundString = self.string
    self.stateString = ""
    self.splitString(self.string, checks = self.split)
    for i in range(len(self.compoundString)):
      if not self.compoundString[i].isdigit():
        self.compoundString = self.compoundString[i:].strip()
        break
    self.readByCharacter(self.compoundString, checks = self.checks, endSetup = self.save)
    self.mass = Measurement.sum([e.mass * Measurement.fromFloat(self.composition[e]) for e in self.composition])

  def split(self, string):
    """Split the string into a compound string and a state string.

    :param string: The string to split.
    :type string: str
    """
    if '(s)' in string or '(l)' in string or '(g)' in string or '(aq)' in string:
      if '(s)' in string:
        self.stateString = '(s)'
      elif '(l)' in string:
        self.stateString = '(l)'
      elif '(g)' in string:
        self.stateString = '(g)'
      else:
        self.stateString = '(aq)'
      self.compoundString = self.string.replace(self.stateString, '').strip()
  
  def checks(self, i, char):
    """Check the character.

    :param i: The index of the character.
    :type i: int
    :param char: The character to check.
    :type char: str
    """
    if char.isalpha():
      if self.parenthesesOn:
          self.subString += char
      else:
        if char.isupper():
          self.save()
          self.element += char
        else:
          self.element += char
    elif char.isdigit():
      if self.parenthesesOn:
        self.subString += char
      else:
        self.number += char
    else:
      if char == '(' or char == '[':
        if not self.parenthesesOn:
          self.save()
          self.parenthesesOn = True
          self.parentheses = ')' if char == '(' else ']'
        else:
          self.subString += char
      elif char == ')' or char == ']':
        if char == self.parentheses:
          self.parenthesesOn = False
        else:
          self.subString += char
      else:
        raise Exception(f"Compound Parser Exception: Unknown character '{char}'")

  def saveElement(self, element, number, multiple = 1):
    """Save the element to the composition.

    :param element: The element to save.
    :type element: str
    :param number: The number of the element.
    :type number: str
    :param multiple: The multiple of the element.
    :type multiple: int
    """
    if Element(element) not in self.composition:
      self.composition[Element(element)] = (int(number) if number else 1) * multiple
    else:
      self.composition[Element(element)] += (int(number) if number else 1) * multiple

  def saveSubcompound(self):
    """Save subcompound to the composition.
    """
    subCompound = Compound(self.subString)
    for element in subCompound.composition:
      self.saveElement(element.string, self.number, multiple = subCompound.composition[element])

  def save(self):
    """Save the element or subcompound to the composition.
    """
    if self.element or self.subString:
      if self.element:
        self.saveElement(self.element, self.number)
      if self.subString:
        self.saveSubcompound()
    self.element = ""
    self.number = ""
    self.subString = ""

  def __str__(self, textOverride = False):
    """Get the string representation of the compound.

    :param textOverride: Whether to override the latex print setting.
    :type textOverride: bool
    :return: The string representation of the compound.
    :rtype: str
    """
    if Compound.latexPrint and not textOverride:
      allCoefficients = re.findall("\d+", self.compoundString)
      finalString = ''
      currentIndex = 0
      for i in allCoefficients:
        ind = self.compoundString[currentIndex:].index(i)
        finalString += self.compoundString[currentIndex:currentIndex + ind]+f'_{{{i}}}'
        currentIndex = currentIndex + ind + len(i)
      if currentIndex < len(self.compoundString):
        finalString += self.compoundString[currentIndex:]
      return finalString + (self.stateString if self.stateString else '')
    return self.compoundString + (self.stateString if self.stateString else '')

  def __eq__(self, other):
    """Check if the compound is equal to another compound.
    
    :param other: The other compound to check.
    :type other: Compound
    :return: Whether the compounds are equal.
    :rtype: bool
    """
    return self.composition == other.composition

  def __ne__(self, other):
    """Check if the compound is not equal to another compound.
    
    :param other: The other compound to check.
    :type other: Compound
    :return: Whether the compounds are not equal.
    :rtype: bool
    """
    return self.composition != other.composition
  
  def massPercentComposition(self):
    """Get the mass percent composition of the compound.

    :return: The mass percent composition of the compound.
    :rtype: list
    """
    return [['Element', 'Mass']]+[[str(i), self.composition[i]*i.mass] for i in self.composition]

  def molePercentComposition(self):
    """Get the mole percent composition of the compound.

    :return: The mole percent composition of the compound.
    :rtype: list
    """
    return [['Element', 'Moles']]+[[str(i), self.composition[i]] for i in self.composition]


from pymeasurement.util.typecheck import typecheck

class Parser:
  """A class to parse strings.

  :param string: The string to parse.
  :type string: str
  :raises TypeError: If string is not a string.
  """
  def __init__(self, string):
    """Parser Constructor
    """
    typecheck(string, str)
    self.string = string
  
  def readByCharacter(self, string, startSetup = None, endSetup = None, checks = None):
    """Read a string by character.

    :param string: The string to read.
    :type string: str
    :param startSetup: A function to run before reading the string.
    :type startSetup: function
    :param endSetup: A function to run after reading the string.
    :type endSetup: function
    :param checks: A function to run on each character.
    :type checks: function
    """
    if startSetup:
      startSetup()
    for i, char in enumerate(string):
      if checks:
        checks(i, char)
    if endSetup:
      endSetup()

  def splitString(self, string, checks = None):
    """Split a string.

	  :param string: The string to split.
    :type string: str
    :param checks: A function to run on each string.
    :type checks: function
	"""
    if checks:
      checks(string)
  
  def __str__(self):
    """Get the string representation of the parser.

    :returns: The string representation of the parser.
    :rtype: str
	"""
    return self.string

  def __repr__(self):
    """Get the string representation of the parser.
    
    :returns: The string representation of the parser.
    :rtype: str
	"""
    return str(self)


from pymeasurement.sigfig import SigFig
import math

class Measurement:
  """
  Measurement
  A class to represent a SigFig sample with a SigFig uncertainty and corresponding units.
  This class can be used to perform calculations with uncertainty propagation.
  Units are also automatically derived through operations with other measurements.

  :param sample: The sample value as a SigFig object or a string.
  :type sample: SigFig or str
  :param precision: The number of significant figures to use when printing the number. If None, the number of significant figures will be automatically determined.
  :type precision: int
  :param uncertainty: The uncertainty of the sample as a SigFig object or a string.
  :type uncertainty: SigFig or str
  :param uncertaintyPercent: If True, the uncertainty will be interpreted as a percentage of the sample value.
  :type uncertaintyPercent: bool
  :param digital: If True, the uncertainty will be automatically determined based on the precision of the device.
  :type digital: bool
  :param analog: If True, the uncertainty will be automatically determined based on the precision of the device.
  :type analog: bool
  :param units: The units of the measurement as a string.
  :type units: str
  :param P: The number of significant figures to use when printing the number. If None, the number of significant figures will be automatically determined.
  :type P: int
  :param U: The uncertainty of the sample as a SigFig object or a string.
  :type U: SigFig or str
  :param UP: If True, the uncertainty will be interpreted as a percentage of the sample value.
  :type UP: bool
  :param D: If True, the uncertainty will be automatically determined based on the precision of the device.
  :type D: bool
  :param A: If True, the uncertainty will be automatically determined based on the precision of the device.
  :type A: bool
  :param UN: The units of the measurement as a string.
  :type UN: str
  """
  def __init__(self, sample, precision=None, uncertainty=None, uncertaintyPercent=False, digital=False, analog=False, units=None, P=None, U=None, UP=False, D=False, A=False, UN=None):
    """
    Measurement Constructor
    """
    if P is not None:
      precision = P
    if U is not None:
      uncertainty = U
    if UP:
      uncertaintyPercent = UP
    if D:
      digital = D
    if A:
      analog = A
    if UN is not None:
      units = UN
    
    if not isinstance(sample, SigFig):
      if '(' in sample or '[' in sample:
        sample, precision = sample.split()
        try:
          precision = int(precision[1:-1]) #Precision in Sig Figs
        except ValueError:
          raise Exception('Measurement Error: Invalid Literal for Sig Fig Precision.')
    if not isinstance(sample, SigFig):
      if precision == float('inf'):
        self.sample = SigFig(sample, constant=True)
      else:
        self.sample = SigFig(sample, sigfigs=precision) 
    else:
      self.sample = sample

    #Automatic Determination of Uncertainty based on Device
    self.uncertainty = None
    if analog:
      self.uncertainty = SigFig(f"5e{self.sample.decimals}", decimals=self.sample.decimals)
    elif digital:
      self.uncertainty = SigFig(f"1e{self.sample.decimals}", decimals=self.sample.decimals)
    
    #Main Override
    if uncertainty is not None:
      if not isinstance(uncertainty, SigFig):
        if '%' in uncertainty:
          uncertaintyPercent = True
          uncertainty = uncertainty.replace('%', '')
      self.uncertainty = SigFig(uncertainty, decimals=self.sample.decimals) if not isinstance(uncertainty, SigFig) else uncertainty
    self.uncertaintyPercent = uncertaintyPercent

    #Chemistry Percent Rules(if <2%, 2 sig figs. Else 1 sig fig)
    if self.uncertainty is not None and self.uncertaintyPercent:
      self.uncertainty = SigFig(str(self.uncertainty.decimalValue), sigfigs=(2 if self.uncertainty < SigFig('2', constant=True) else 1))

    #Determine Units
    #Will use units class to allow for conversions later.
    self.units = units
    self.nUnits = [i.strip('() ') for i in self.units.split('*')] if self.units is not None else []
    self.dUnits = []
    if self.units is not None:
      if '/' in units:
        nUnitsStr, dUnitsStr = self.units.split('/')
        nUnitsStr = nUnitsStr.strip('() ')
        dUnitsStr = dUnitsStr.strip('() ')
        self.nUnits = [i.strip('() ') for i in nUnitsStr.split('*')]
        self.dUnits = [i.strip('() ') for i in dUnitsStr.split('*')]
        if '^' in nUnitsStr:
          newNUnits = []
          for i in self.nUnits:
            if '^' in i:
              i, repeat = i.split('^')
              for n in range(int(repeat)-1):
                newNUnits.append(i)
            newNUnits.append(i)
          self.nUnits = newNUnits
        if '^' in dUnitsStr:
          newDUnits = []
          for i in self.dUnits:
            if '^' in i:
              i, repeat = i.split('^')
              for n in range(int(repeat)-1):
                newDUnits.append(i)
            newDUnits.append(i)
          self.dUnits = newDUnits
    newNUnits = self.nUnits
    newDUnits = self.dUnits
    for i in self.nUnits:
      if i in newDUnits:
        newNUnits.remove(i)
        newDUnits.remove(i)
    #Sort units alphabetically
    self.nUnits=sorted(newNUnits)
    self.dUnits=sorted(newDUnits)
    #Reformat units string
    self.units = Measurement.formatUnits(self.nUnits, self.dUnits)

  def fromStr(string):
    """Creates a Measurement object from a string.
    The string must be in the form of a number, uncertainty, and units.
    The uncertainty can be in the form of a percentage or a number.
    A 'a' or 'd' can be used to indicate an analog or digital device for automatic uncertainty determination.

    :param string: The string to create the Measurement object from.
    :type string: str
    :return: The Measurement object created from the string.
    :rtype: Measurement
    """
    sample = string.strip()
    uncertainty = None
    units = None
    values = sample.split()
    if '+/-' in sample or '+-' in sample:
      if len(values) == 3:
        if '+/-' in sample:
          sample, uncertainty = string.split('+/-')
        elif '+-' in sample:
          sample, uncertainty = string.split('+-')
      elif len(values) >= 4:
        sample, _, uncertainty, *units = values
        if len(units) == 2: # Assuming form will mol H2O
          if '_' not in units[1]:
            from pymeasurement.util.chem.compound import Compound
            units[1] = str(Compound(units[1])) # Format Compound String
        units = ' '.join(units)
    else:
      if len(values) >= 2:
        sample, *units = values
        if len(units) == 2: # Assuming form will mol H2O
          if '_' not in units[1]:
            from pymeasurement.util.chem.compound import Compound
            units[1] = str(Compound(units[1])) # Format Compound String
        units = ' '.join(units)
    precision = None
    digital = False
    analog = False
    if uncertainty is None:
      if 'c' in sample:
        precision = float('inf')
        sample = sample.replace('c', '')
      elif 'd' in sample:
        digital = True
        sample = sample.replace('d', '')
      elif 'a' in sample:
        analog = True
        sample = sample.replace('a', '')
    return Measurement(sample.strip(), precision=precision, uncertainty=(uncertainty.strip() if isinstance(uncertainty, str) else uncertainty), digital=digital, analog=analog, units=units)

  def fromFloat(f, units=''): #Assume float is a constant with infinite precision and no uncertainty.
    """
    Creates a Measurement constant from a float.

    :param f: The float to create the Measurement constant from.
    :type f: float
    :param units: The units of the Measurement constant.
    :type units: str
    :return: The Measurement constant created from the float.
    :rtype: Measurement
    """
    return Measurement.fromStr(f'{f}c {units}')
  
  def toAbsolute(self):
    """
    Converts the uncertainty to an absolute value. Note that this mutates the object.

    :return: The Measurement object with the uncertainty converted to an absolute value.
    :rtype: Measurement
    """
    if self.uncertaintyPercent and isinstance(self.uncertainty, SigFig):
      self.uncertaintyPercent = False
      self.uncertainty *= (self.sample / SigFig('100', constant=True)).abs()
      self.uncertainty = SigFig(str(self.uncertainty.decimalValue), decimals=self.sample.decimals)
    return self

  def toPercent(self):
    """
    Converts the uncertainty to a percentage. Note that this mutates the object.

    :return: The Measurement object with the uncertainty converted to a percentage.
    :rtype: Measurement
    """
    if not self.uncertaintyPercent and isinstance(self.uncertainty, SigFig):
      self.uncertaintyPercent = True
      self.uncertainty = SigFig(self.uncertainty.value, constant=True) * SigFig((SigFig('100', constant=True) / self.sample).abs().value, constant=True)
      self.uncertainty = SigFig(str(self.uncertainty.decimalValue), sigfigs=(2 if self.uncertainty < SigFig('2', constant=True) else 1))
    return self

  def absolute(m):
    """
    Returns a copy of the Measurement with the uncertainty converted to an absolute value.

    :return: A copy of the Measurement with the uncertainty converted to an absolute value.
    :rtype: Measurement
    """
    return m.deepCopy().toAbsolute()

  def percent(m):
    """
    Returns a copy of the Measurement with the uncertainty converted to a percentage.

    :return: A copy of the Measurement with the uncertainty converted to a percentage.
    :rtype: Measurement
    """
    return m.deepCopy().toPercent()
  
  def deepCopy(self):
    """
    Returns a deep copy of the Measurement object.

    :return: A deep copy of the Measurement object.
    :rtype: Measurement
    """
    return Measurement(self.sample.deepCopy(), uncertainty = self.uncertainty.deepCopy() if self.uncertainty is not None else None, uncertaintyPercent = self.uncertaintyPercent, units=self.units)
  
  def __str__(self):
    """
    Returns a string representation of the Measurement object.

    :return: A string representation of the Measurement object.
    :rtype: str
    """
    return str(self.sample) + (f' +/- {self.uncertainty}' + ('%' if self.uncertaintyPercent else '') if isinstance(self.uncertainty, SigFig) else '') + (f' {Measurement.formatUnits(self.nUnits, self.dUnits)}' if self.units is not None else '')

  def __repr__(self):
    """
    Returns a string representation of the Measurement object.

    :return: A string representation of the Measurement object.
    :rtype: str
    """
    return str(self)

  def multUnits(nUnits1, dUnits1, nUnits2, dUnits2):
    """
    Multiplies two sets of units.

    :param nUnits1: The numerator units of the first set.
    :type nUnits1: list
    :param dUnits1: The denominator units of the first set.
    :type dUnits1: list
    :param nUnits2: The numerator units of the second set.
    :type nUnits2: list
    :param dUnits2: The denominator units of the second set.
    :type dUnits2: list
    :return: The multiplied units.
    :rtype: tuple
    """
    nUnits = nUnits1 + nUnits2
    dUnits = dUnits1 + dUnits2
    newNUnits = nUnits
    newDUnits = dUnits
    for i in nUnits:
      if i in newDUnits:
        newNUnits.remove(i)
        newDUnits.remove(i)
    return (sorted(newNUnits), sorted(newDUnits))

  def formatUnits(nUnits, dUnits):
    """
    Formats a set of units into a string.

    :param nUnits: The numerator units.
    :type nUnits: list
    :param dUnits: The denominator units.
    :type dUnits: list
    :return: The formatted units.
    :rtype: str
    """
    combinedNUnits = sorted([i if nUnits.count(i) == 1 else f'{i}^{nUnits.count(i)}' for i in set(nUnits)])
    combinedDUnits = sorted([i if dUnits.count(i) == 1 else f'{i}^{dUnits.count(i)}' for i in set(dUnits)])
    nUnitsStr = '1' if not nUnits else '*'.join(combinedNUnits)
    dUnitsStr = '' if not dUnits else '*'.join(combinedDUnits)
    if len(combinedNUnits) > 1:
      nUnitsStr = '(' + nUnitsStr + ')'
    if len(combinedDUnits) > 1:
      dUnitsStr = '(' + dUnitsStr + ')'
    return None if nUnitsStr == '1' and dUnitsStr == '' else nUnitsStr + (f'/{dUnitsStr}' if dUnitsStr else '')

  def __eq__(self, other):
    """
    Returns True if the two Measurement objects are equal.

    :param other: The Measurement object to compare.
    :type other: Measurement
    :returns: True if the two Measurement objects are equal.
    :rtype: bool
    """
    return self.sample == other.sample

  def __lt__(self, other):
    """
    Returns True if the first Measurement object is less than the second.

    :param other: The Measurement object to compare.
    :type other: Measurement
    :returns: True if the first Measurement object is less than the second.
    :rtype: bool
    """
    return self.sample < other.sample

  def __gt__(self, other):
    """
    Returns True if the first Measurement object is greater than the second.

    :param other: The Measurement object to compare.
    :type other: Measurement
    :returns: True if the first Measurement object is greater than the second.
    :rtype: bool
    """
    return self.sample > other.sample

  def __le__(self, other):
    """
    Returns True if the first Measurement object is less than or equal to the second.

    :param other: The Measurement object to compare.
    :type other: Measurement
    :returns: True if the first Measurement object is less than or equal to the second.
    :rtype: bool
    """
    return self < other or self == other

  def __ge__(self, other):
    """
    Returns True if the first Measurement object is greater than or equal to the second.

    :param other: The Measurement object to compare.
    :type other: Measurement
    :returns: True if the first Measurement object is greater than or equal to the second.
    :rtype: bool
    """
    return self > other or self == other
  
  def __neg__(self):
    """
    Returns the negation of the Measurement object.

    :returns: The negation of the Measurement object.
    :rtype: Measurement
    """
    neg = self.deepCopy()
    neg.sample = -self.sample
    return neg

  def __add__(self, other):
    """
    Returns the sum of the two Measurement objects.

    :param other: The Measurement object to add.
    :type other: Measurement
    :returns: The sum of the two Measurement objects.
    :rtype: Measurement
    """
    if self.nUnits != other.nUnits or self.dUnits != other.dUnits:
      raise Exception(f'Measurement Error: Cannot add {self} and {other} with different units.')
    uSum = SigFig('0', constant=True)
    uncertainties = [Measurement.absolute(i).uncertainty for i in [self, other] if i.uncertainty is not None]
    for u in uncertainties:
      uSum += u
    return Measurement(self.sample + other.sample, uncertainty=uSum if uncertainties else None, units=self.units)
  
  def __radd__(self, other):
    """
    Returns the sum of the two Measurement objects.

    :param other: The Measurement object to add to.
    :type other: Measurement
    :returns: The sum of the two Measurement objects.
    :rtype: Measurement
    """
    return self + other
  
  def __sub__(self, other):
    """
    Returns the difference of the two Measurement objects.

    :param other: The Measurement object to subtract.
    :type other: Measurement
    :returns: The difference of the two Measurement objects.
    :rtype: Measurement
    """
    return -other + self

  def __rsub__(self, other):
    """
    Returns the difference of the two Measurement objects.

    :param other: The Measurement object to subtract from.
    :type other: Measurement
    :returns: The difference of the two Measurement objects.
    :rtype: Measurement
    """
    return -self + other

  def __mul__(self, other):
    """
    Returns the product of the two Measurement objects.

    :param other: The Measurement object to multiply by.
    :type other: Measurement
    :returns: The product of the two Measurement objects.
    :rtype: Measurement
    """
    if isinstance(other, float) or isinstance(other, int):
      other = Measurement.fromFloat(other)
    uSum = SigFig('0', constant=True)
    uncertainties = [Measurement.percent(i).uncertainty for i in [self, other] if i.uncertainty is not None]
    for u in uncertainties:
      uSum += u
    nUnits, dUnits = Measurement.multUnits(self.nUnits, self.dUnits, other.nUnits, other.dUnits)
    return Measurement(self.sample * other.sample, uncertainty=uSum if uncertainties else None, uncertaintyPercent=True, units=Measurement.formatUnits(nUnits, dUnits))
  
  def __rmul__(self, other):
    """
    Returns the product of the two Measurement objects.

    :param other: The Measurement object to multiply by.
    :type other: Measurement
    :returns: The product of the two Measurement objects.
    :rtype: Measurement
    """
    return self * other

  def __truediv__(self, other):
    """
    Returns the quotient of the two Measurement objects.

    :param other: The Measurement object to divide by.
    :type other: Measurement
    :returns: The quotient of the two Measurement objects.
    :rtype: Measurement
    """
    if isinstance(other, float) or isinstance(other, int):
      other = Measurement.fromFloat(other)
    uSum = SigFig('0', constant=True)
    uncertainties = [Measurement.percent(i).uncertainty for i in [self, other] if i.uncertainty is not None]
    for u in uncertainties:
      uSum += u
    nUnits, dUnits = Measurement.multUnits(self.nUnits, self.dUnits, other.dUnits, other.nUnits)
    return Measurement(self.sample / other.sample, uncertainty=uSum if uncertainties else None, uncertaintyPercent=True, units=Measurement.formatUnits(nUnits, dUnits))
    
  def __rtruediv__(self, other):
    """
    Returns the quotient of the two Measurement objects.

    :param other: The Measurement object to divide by.
    :type other: Measurement
    :returns: The quotient of the two Measurement objects.
    :rtype: Measurement
    """
    return other / self

  def __pow__(self, integer):
    """
    Returns the Measurement object raised to the given integer power.

    :param integer: The integer power to raise the Measurement object to.
    :type integer: int
    :returns: The Measurement object raised to the given integer power.
    :rtype: Measurement
    """
    product = Measurement('1', precision=float('inf'))
    for i in range(integer):
      product *= self
    return product

  def sum(measurements):
    """
    Returns the sum of the given list of Measurement objects.

    :param measurements: The list of Measurement objects.
    :type measurements: list
    :returns: The sum of the given list of Measurement objects.
    :rtype: Measurement
    """
    if not measurements:
      return Measurement.fromStr('0c')
    s = measurements[0]
    for i in measurements[1:]:
      s += i
    return s

  def max(measurements):
    """
    Returns the maximum of the given list of Measurement objects.

    :param measurements: The list of Measurement objects.
    :type measurements: list
    :returns: The maximum of the given list of Measurement objects.
    :rtype: Measurement
    """
    m = measurements[0]
    for i in measurements[1:]:
      if m < i:
        m = i
    return m

  def min(measurements):
    """
    Returns the minimum of the given list of Measurement objects.

    :param measurements: The list of Measurement objects.
    :type measurements: list
    :returns: The minimum of the given list of Measurement objects.
    :rtype: Measurement
    """
    m = measurements[0]
    for i in measurements[1:]:
      if m > i:
        m = i
    return m

  def average(measurements):
    """
    Returns the average of the given list of Measurement objects. Uses (max - min) / (2 * sqrt(n)) as the uncertainty.

    :param measurements: The list of Measurement objects.
    :type measurements: list
    :returns: The average of the given list of Measurement objects.
    :rtype: Measurement
    """
    avg_sample = Measurement.sum(measurements) / len(measurements)
    avg_uncertainty = (Measurement.max(measurements) - Measurement.min(measurements)).sample / (2 * math.sqrt(len(measurements)))
    return Measurement(avg_sample.sample, uncertainty=avg_uncertainty.value, units=avg_sample.units)

  def convert(sample, uncertainty=None, uncertaintyPercent=False, units='', analog=False, digital=False, constant=False, u=None, up=False, a=False, d=False, un='', decimals=None):
    """
    Returns a Measurement object with the given sample, uncertainty, and units.
    
    :param sample: The sample of the Measurement object.
    :type sample: float or str
    :param uncertainty: The uncertainty of the Measurement object.
    :type uncertainty: float or str
    :param uncertaintyPercent: Whether the uncertainty is a percent.
    :type uncertaintyPercent: bool
    :param units: The units of the Measurement object.
    :type units: str
    :param analog: Whether the Measurement object is analog.
    :type analog: bool
    :param digital: Whether the Measurement object is digital.
    :type digital: bool
    :param constant: Whether the Measurement object is constant.
    :type constant: bool
    :param u: The uncertainty of the Measurement object.
    :type u: float or str
    :param up: Whether the uncertainty is a percent.
    :type up: bool
    :param a: Whether the Measurement object is analog.
    :type a: bool
    :param d: Whether the Measurement object is digital.
    :type d: bool
    :param un: The units of the Measurement object.
    :type un: str
    :param decimals: The number of decimals to round to.
    :type decimals: int
    """
    if u is not None:
      uncertainty = u
    if up:
      uncertaintyPercent = up
    if a:
      analog = a
    if d:
      digital = d
    if un:
      units = un

    return Measurement(SigFig(str(sample), decimals=-decimals if decimals is not None else None), uncertaintyPercent=uncertaintyPercent, uncertainty=str(uncertainty) if uncertainty is not None else None, precision=float('inf') if constant else None, units=units, analog=analog, digital=digital)
  
  def importColumn(column, uncertaintyColumn=None, df=None, **kwargs):
    """
    Convert a numeric Pandas DataFrame column to Measurement objects.
    
    :param column: The numeric Pandas DataFrame column.
    :type column: pandas.core.series.Series
    :param kwargs: Keyword arguments to pass to Measurement.convert.
    :type kwargs: dict
    """
    if uncertaintyColumn is None:
      return column.apply(Measurement.convert, **kwargs)
    else:
      return df.apply(lambda x: Measurement.convert(x[column.name], u=x[uncertaintyColumn.name], **kwargs), axis=1)

  def exportColumn(savedf, column, addUncertainty=True, asPercent=True):
    """
    Convert a Measurement Pandas DataFrame column to numeric values.

    :param savedf: The Pandas DataFrame to save to.
    :type savedf: pandas.core.frame.DataFrame
    :param column: The Measurement Pandas DataFrame column.
    :type column: pandas.core.series.Series
    :param addUncertainty: Whether to add the uncertainty to the DataFrame.
    :type addUncertainty: bool
    :param asPercent: Whether to add the uncertainty as a percent.
    :type asPercent: bool
    """
    savedf[column.name] = column.apply(lambda x: x.sample)
    if addUncertainty:
      label, units = (' ('.join(column.name.split(' (')[:-1]), ' (' + column.name.split(' (')[-1]) if ' (' in column.name else (column.name, '')
      if asPercent:
        savedf.insert(savedf.columns.get_loc(column.name) + 1, f'{label} Percent Uncertainty (%)', column.apply(lambda x: x.percent().uncertainty))
      else:
        savedf.insert(savedf.columns.get_loc(column.name) + 1, f'{label} Absolute Uncertainty{units}', column.apply(lambda x: x.absolute().uncertainty))



import unittest
import sys
import os
sys.path.insert(0, os.path.join(os.path.dirname(__file__), '..', 'src'))
from pymeasurement import Measurement
from pymeasurement.sigfig import SigFig

class TestMeasurement(unittest.TestCase):
    def test_create_measurement(self):
        m = Measurement("2.0", uncertainty="0.13", units="m")
        self.assertEqual(m.sample, SigFig("2.0"))
        self.assertEqual(m.uncertainty, SigFig("0.1"))
        self.assertEqual(m.units, "m")
        self.assertEqual(str(m), "2.0 +/- 0.1 m")

    def test_create_measurement_with_sigfigs(self):
        m = Measurement("2.0", uncertainty="0.13", units="m", precision=3)
        self.assertEqual(m.sample, SigFig("2.00"))
        self.assertEqual(m.uncertainty, SigFig("0.13"))
        self.assertEqual(m.units, "m")
        self.assertEqual(str(m), "2.00 +/- 0.13 m")

    def test_create_measurement_from_str(self):
        m = Measurement.fromStr("2.0 +/- 0.13 m")
        self.assertEqual(m.sample, SigFig("2.0"))
        self.assertEqual(m.uncertainty, SigFig("0.1"))
        self.assertEqual(m.units, "m")
        self.assertEqual(str(m), "2.0 +/- 0.1 m")

    def test_create_measurement_from_float(self):
        m = Measurement.fromFloat(3.14)
        self.assertEqual(m.sample, SigFig("3.14", constant=True))
        self.assertEqual(m.uncertainty, None)
        self.assertEqual(m.sample.sigfigs, float("inf"))
        self.assertEqual(str(m), "3.14")

    def test_add_measurements(self):
        m1 = Measurement("2.0", uncertainty="0.13", units="m")
        m2 = Measurement("3.0", uncertainty="0.1", units="m")
        m3 = m1 + m2
        self.assertEqual(m3.sample, SigFig("5.0"))
        self.assertEqual(m3.uncertainty, SigFig("0.2"))
        self.assertEqual(m3.units, "m")
        self.assertEqual(str(m3), "5.0 +/- 0.2 m")
    
    def test_subtract_measurments(self):
        m1 = Measurement("2.0", uncertainty="0.13", units="m")
        m2 = Measurement("3.0", uncertainty="0.1", units="m")
        m3 = m1 - m2
        self.assertEqual(m3.sample, SigFig("-1.0"))
        self.assertEqual(m3.uncertainty, SigFig("0.2"))
        self.assertEqual(m3.units, "m")
        self.assertEqual(str(m3), "-1.0 +/- 0.2 m")

    def test_multiply_measurements(self):
        m1 = Measurement("2.0", uncertainty="0.13", units="m")
        m2 = Measurement("3.0", uncertainty="0.1", units="m")
        m3 = m1 * m2
        self.assertEqual(m3.sample, SigFig("6.0"))
        self.assertEqual(m3.absolute().uncertainty, SigFig("0.6"))
        self.assertEqual(m3.units, "m^2")
        self.assertEqual(str(m3), "6.0 +/- 1E+1% m^2")

    def test_divide_measurements(self):
        m1 = Measurement("2.0", uncertainty="0.13", units="m")
        m2 = Measurement("3.0", uncertainty="0.1", units="m")
        m3 = m1 / m2
        self.assertEqual(m3.sample, SigFig("0.67"))
        self.assertEqual(m3.uncertainty, SigFig("10", sigfigs=1))
        self.assertEqual(m3.units, None)
        self.assertEqual(str(m3), "0.67 +/- 1E+1%")

    def test_add_measurement_to_scalar(self):
        m1 = Measurement("2.0", uncertainty="0.13", units="m")
        m2 = m1 + Measurement.fromFloat(2, units="m")
        self.assertEqual(m2.sample, SigFig("4.0"))
        self.assertEqual(m2.uncertainty, SigFig("0.1"))
        self.assertEqual(m2.units, "m")
        self.assertEqual(str(m2), "4.0 +/- 0.1 m")

    def test_subtract_scalar_from_measurement(self):
        m1 = Measurement("2.0", uncertainty="0.13", units="m")
        m2 = m1 - Measurement.fromFloat(2, units="m")
        self.assertEqual(m2.sample, SigFig("0.0"))
        self.assertEqual(m2.uncertainty, SigFig("0.1"))
        self.assertEqual(m2.units, "m")
        self.assertEqual(str(m2), "0.0 +/- 0.1 m")

    def test_divide_measurement_by_scalar(self):
        m1 = Measurement("2.0", uncertainty="0.13", units="m")
        m2 = m1 / 2
        self.assertEqual(m2.sample, SigFig("1.0"))
        self.assertEqual(m2.uncertainty, SigFig("6"))
        self.assertEqual(m2.units, "m")
        self.assertEqual(str(m2), "1.0 +/- 6% m")

    def test_multiply_measurement_by_scalar(self):
        m1 = Measurement("2.0", uncertainty="0.13", units="m")
        m2 = m1 * 2
        self.assertEqual(m2.sample, SigFig("4.0"))
        self.assertEqual(m2.uncertainty, SigFig("6"))
        self.assertEqual(m2.units, "m")
        self.assertEqual(str(m2), "4.0 +/- 6% m")

    def test_square_measurement(self):
        m1 = Measurement("2.0", uncertainty="0.13", units="m")
        m2 = m1 ** 2
        self.assertEqual(m2.sample, SigFig("4.0"))
        self.assertEqual(m2.uncertainty, SigFig("1E+1"))
        self.assertEqual(m2.units, "m^2")
        self.assertEqual(str(m2), "4.0 +/- 1E+1% m^2")

    def test_absolute_uncertainty(self):
        m = Measurement.fromStr("2.0 +/- 0.13 m")
        self.assertEqual(m.absolute().uncertainty, SigFig("0.1"))
        self.assertEqual(m.absolute().units, "m")
        self.assertEqual(str(m.absolute()), "2.0 +/- 0.1 m")

    def test_percent_uncertainty(self):
        m = Measurement.fromStr("2.0 +/- 0.13 m")
        self.assertEqual(m.percent().uncertainty, SigFig("6", sigfigs=1))
        self.assertEqual(m.percent().units, "m")
        self.assertEqual(str(m.percent()), "2.0 +/- 6% m")

    def test_sum_measurements(self):
        m1 = Measurement.fromStr("2.0 +/- 0.13 m")
        m2 = Measurement.fromStr("3.0 +/- 0.1 m")
        m3 = Measurement.fromStr("4.0 +/- 0.1 m")
        m4 = Measurement.sum([m1, m2, m3])
        self.assertEqual(m4.sample, SigFig("9.0"))
        self.assertEqual(m4.uncertainty, SigFig("0.3"))
        self.assertEqual(m4.units, "m")
        self.assertEqual(str(m4), "9.0 +/- 0.3 m")

    def test_max_measurements(self):
        m1 = Measurement.fromStr("2.0 +/- 0.13 m")
        m2 = Measurement.fromStr("3.0 +/- 0.1 m")
        m3 = Measurement.fromStr("4.0 +/- 0.1 m")
        m4 = max([m1, m2, m3])
        self.assertEqual(m4.sample, SigFig("4.0"))
        self.assertEqual(m4.uncertainty, SigFig("0.1"))
        self.assertEqual(m4.units, "m")
        self.assertEqual(str(m4), "4.0 +/- 0.1 m")

    def test_min_measurements(self):
        m1 = Measurement.fromStr("2.0 +/- 0.13 m")
        m2 = Measurement.fromStr("3.0 +/- 0.1 m")
        m3 = Measurement.fromStr("4.0 +/- 0.1 m")
        m4 = min([m1, m2, m3])
        self.assertEqual(m4.sample, SigFig("2.0"))
        self.assertEqual(m4.uncertainty, SigFig("0.1"))
        self.assertEqual(m4.units, "m")
        self.assertEqual(str(m4), "2.0 +/- 0.1 m")    

    def test_average_measurements(self):
        m1 = Measurement.fromStr("2.0 +/- 0.13 m")
        m2 = Measurement.fromStr("3.0 +/- 0.1 m")
        m3 = Measurement.fromStr("4.0 +/- 0.1 m")
        m4 = Measurement.average([m1, m2, m3])
        self.assertEqual(m4.sample, SigFig("3.0"))
        self.assertEqual(m4.uncertainty, SigFig("0.6"))
        self.assertEqual(m4.units, "m")
        self.assertEqual(str(m4), "3.0 +/- 0.6 m")